/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "conf.h"

//-----------------------------------------------------------------------------
// Configure the program

bool TCVEMConf::Configure(string &err){

    err.clear();

    // Reding the conf file to the memory
    if(!ReadFile(MAIN_CONF)){
        err = "Can't read configuration file";
        return false;
        }

    // Parsing the conf file (forming a DOM)
    if(!Parse()){
        err = "Wrong configuration file format";
        return false;
        }

    // Applying the configuration according to a DOM of the conf file
    if(!ParseConfDOM(err)){
        err += err.empty() ? "Configuration " : ", configuration ";
        err += "can't be applied";
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Parse the database section

bool TCVEMConf::ParseDBSec(string &err){

    for(auto &[k, v] : db_connection){
        if(!ReadDOMParam(v, DB_CON_SEC, k)){
            err = "Missing some of the DB connection params";
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Parse the common section

bool TCVEMConf::ParseCommonSec(string &err){

    for(auto &[k, v] : common){
        if(!ReadDOMParam(v, COMMON_SEC, k)){
            err = "Some of the params are missing in the common "
                "section of the configuration file";
            return false;
            }
        if(common_params_file_paths.find(k) != common_params_file_paths.end()){
            v = FormatFilePath(v, IS_DIR);
            }
        }

    vector<string> optional_params;

    // Reading optional parameters
    if(ReadVParams(optional_params, COMMON_SEC, "excluded_vulsrc")){
        for(const auto &vulsrc : optional_params){
            // Checking if a source with a given name is available
            if(available_vulsrc.find(vulsrc) != available_vulsrc.end()){
                excluded_vulsrc.push_back(vulsrc);
                }
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Apply configuration that was read from the conf file

bool TCVEMConf::ParseConfDOM(string &err){

    err.clear();

    const map<bool, vector<string>> branch_params = {
        {true, {"branch", "base_dir", "srpms_dir", "bin_list"}},
        {false, {"rpms_dirs"}}
        };

    for(auto sec = AllSec().rbegin(); sec != AllSec().rend(); sec++){

        const auto &sec_name = sec->Name();

        // If this is the database section
        if(sec_name == DB_CON_SEC){
            if(!ParseDBSec(err)){
                return false;
                }
            }
        // If this is the common section
        else if(sec_name == COMMON_SEC){
            if(!ParseCommonSec(err)){
                return false;
                }
            }
        // Rest of the sections considered as branch descriptions
        else{
            // Skipping repetitions
            if(branches.find(sec_name) != branches.end()){
                continue;
                }
            for(const auto &[is_single_assgn, params] : branch_params){
                for(const auto &param : params){
                    branches[sec_name][param] = is_single_assgn ?
                        sec->VarSingleAssgn(param) :
                        Join(sec->VarAssgn(param), "//");
                    }
                }
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
