/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    <sstream>

#include    "database.h"

//-----------------------------------------------------------------------------
// Create DB connection

bool TDatabase::Connect(string &err, const bool force){

    err.clear();

    if(con && con->isValid() && !force){
        return true;
        }

    driver = get_driver_instance();

    try{
        con = driver->connect(host, user, password);
        }
    catch(...){
        err = "Can't connect to DB";
        return false;
        }

    // Creating the DB if it doesn't exist
    if(!DBExists(err) && !CreateDB(err)){
        err = "Can't create DB" + (!err.empty() ? "; " + err : err);
        return false;
        }

    con->setSchema(database);

    if(res){
        delete res;
        res = NULL;
        }

    if(stmt){
        delete stmt;
        stmt = NULL;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Setup the DB connection params

#define DB_CON_HOST     "host"
#define DB_CON_USER     "user"
#define DB_CON_PASSWORD "password"
#define DB_CON_DATABASE "database"

bool TDatabase::Configure(const map<string, string> &db_connection,
        string &err){

    err.clear();

    const array<string, 4> params =
		{DB_CON_HOST, DB_CON_USER, DB_CON_PASSWORD, DB_CON_DATABASE};

    // Checking given connection params
    for(const auto &param : params){

        const auto it = db_connection.find(param);

        if(it == db_connection.end()){
            err = "\"" + param + "\" parameter is missing";
            return false;
            }

        const auto &value = it->second;

        if(param != DB_CON_PASSWORD && value.empty()){
            err = "\"" + param + "\" parameter can't be empty";
            return false;
            }
        }

    host     = db_connection.at(DB_CON_HOST);
    user     = db_connection.at(DB_CON_USER);
    password = db_connection.at(DB_CON_PASSWORD);
    database = db_connection.at(DB_CON_DATABASE);

    return true;
    }

#undef DB_CON_HOST
#undef DB_CON_USER
#undef DB_CON_PASSWORD
#undef DB_CON_DATABASE

//-----------------------------------------------------------------------------
// Send noreturn query to the DB

bool TDatabase::Execute(const string &query, string &err){

    err.clear();

    try{
        if(Connect(err)){
            stmt = con->createStatement();
            stmt->execute(query);
            }
        else{
            return false;
            }
        }
    catch(sql::SQLException &e){
        err = FormatSQLException(e);
        return false;
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Send given query to the DB and return the result

sql::ResultSet* TDatabase::GetResult(const string &query, string &err){

    err.clear();

    try{
        if(Connect(err)){
            stmt = con->createStatement();
            res  = stmt->executeQuery(query);
            }
        else{
            return NULL;
            }
        }
    catch(sql::SQLException &e){
        err = FormatSQLException(e);
        return NULL;
        }

    return res;
    }

//-----------------------------------------------------------------------------
// Fetch all results into matrix of strings

vector<vector<string>> TDatabase::FetchAll(sql::ResultSet *res,
        const size_t n_cols){

    vector<vector<string>> res_matrix;

    if(!res){
        return {};
        }

    while(res->next()){
        vector<string> res_vec;
        for(size_t i = 1; i <= n_cols; i++){
            res_vec.push_back(res->getString(i));
            }
        res_matrix.push_back(res_vec);
        }

    return res_matrix;
    }

//-----------------------------------------------------------------------------
// Check DB or table existence

bool TDatabase::ExistenceCheck(const string &query, string &err){

    err.clear();

    Connect(err);

    if(con && con->isValid() && !query.empty()){
        try{
            stmt = con->createStatement();
            res = stmt->executeQuery(query);
            while(res->next()){
                return true;
                }
            }
        catch(sql::SQLException &e){
            err = FormatSQLException(e);
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Check existence of the DB

bool TDatabase::DBExists(string &err){

    const string query = "SELECT SCHEMA_NAME FROM INFORMATION_SCHEMA.SCHEMATA "
        "WHERE SCHEMA_NAME = '" + database + "'";

    return database.empty() ? false: ExistenceCheck(query, err);
    }

//-----------------------------------------------------------------------------
// Get names of all tables of the DB

bool TDatabase::GetTableNames(vector<string> &table_names, string &err){

    err.clear();
    table_names.clear();

    Connect(err);

    if(con && con->isValid()){
        try{
            stmt = con->createStatement();
            res = stmt->executeQuery("SHOW TABLES");
            while(res->next()){
                table_names.push_back(res->getString(1));
                }
            }
        catch(sql::SQLException &e){
            err = "Can't show " + database + " tables";
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Check table existence in the DB

bool TDatabase::TableExists(const string &name, string &err){

    const string query = "SHOW TABLES LIKE '" + name + "'";

    return ExistenceCheck(query, err);
    }

//-----------------------------------------------------------------------------
// Check that table is not empty

bool TDatabase::TableNotEmpty(const string &name, string &err){

    err.clear();

    Connect(err);

    if(con && con->isValid() && !name.empty() && TableExists(name, err)){
        try{
            stmt = con->createStatement();
            res = stmt->executeQuery("SELECT COUNT(*) FROM " + name);
            if(res->next()){
                if(res->getString(1) != "0"){
                    return true;
                    }
                }
            }
        catch(sql::SQLException &e){
            err = FormatSQLException(e);
            }
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Create the DB

bool TDatabase::CreateDB(string &err){

    const string query = "CREATE DATABASE " + database;

    return Execute(query, err);
    }

//-----------------------------------------------------------------------------
// Use the DB

bool TDatabase::UseDB(string &err){

    const string query = "USE " + database;

    return Execute(query, err);
    }

//-----------------------------------------------------------------------------
// Delete table with a given name

bool TDatabase::DropTable(const string &table, string &err){

    const string query = "DROP TABLE IF EXISTS " + table;

    return Execute(query, err);
    }

//-----------------------------------------------------------------------------
// Copy structure of the table but not any of it's rows

bool TDatabase::CopyTable(const string &table, const string &new_table,
        string &err){

    const string query = "CREATE TABLE " + new_table + " LIKE " + table;

    return Execute(query, err);
    }

//-----------------------------------------------------------------------------
// Drop tables that have specified names

bool TDatabase::DropChildTables(const vector<string> &tables, string &err){

    for(const auto &table : tables){
        if(!DropTable(table, err)){
            return false;
            }
        }

    return true;
    }

//-----------------------------------------------------------------------------
// Push table name into vector of table names if there is a table with this
// name in the DB

bool TDatabase::PushTable(vector<string> &tables, const string &table,
        string &err){

    err.clear();

    if(!TableExists(table, err)){
        if(!err.empty()){
            err = "Checking existence of table \"" + table + "\": " + err;
            }
        return false;
        }

    PushBack(tables, table, 2);

    return true;
    }

//-----------------------------------------------------------------------------
// Format given SQL exception message

string TDatabase::FormatSQLException(const sql::SQLException &e) const{

    ostringstream oss;

    oss << "SQLException in " << __FILE__ << " (" << __FUNCTION__ << ") "
        << "on line " << __LINE__ << "; " << e.what() << " ("
        << "MySQL error code: " << e.getErrorCode() << ", "
        << "SQLState: " << e.getSQLState() << ")";

    return oss.str();
    }

//-----------------------------------------------------------------------------
// Memory deallocation

void TDatabase::Clear(void){

    if(res){
        delete res;
        res = NULL;
        }

    if(stmt){
        delete stmt;
        stmt = NULL;
        }

    if(con){
        delete con;
        con = NULL;
        }
    }

//-----------------------------------------------------------------------------
