/*
    cve-manager : CVE management tool
    Copyright (C) 2017-2024 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "vul.h"

//-----------------------------------------------------------------------------
// Identify format of a given line

// Format 1 is a listing of vulnerability IDs placed right after a variation
// of a keyword ('fix', 'fixed' etc.)

// Format 2 is a multiline listing of vulnerability IDs, each line starts with
// the '+' or the '*' symbol followed by a listing of vulnerability IDs, the
// keyword is the same

int IdentifyFormat(const int current_fmt, string &line){

    if(current_fmt == 0){

        smatch match;

        if(!regex_search(line, match, fixes_regex)){
            return (-1) * abs(current_fmt);
            }

        const auto &pre  = match[1].str();
        const auto &post = match[5].str();

        if(pre.empty() || regex_search(pre, negation_regex) ||
                (!post.empty() && regex_search(post, negation_regex))){
            return (-1) * abs(current_fmt);
            }

        if(pre.front() == '-' && (post.empty() || post.back() == ':')){
            return -2;
            }

        line.clear();

        array<string, 2> pre_and_post = {pre, post};
        for(const auto &piece : pre_and_post){
            if(!piece.empty()){
                smatch search;
                if(regex_search(piece, search, vul_id_regex)){
                    line += piece.substr(search.position());
                    }
                }
            }

        return line.empty() ? -1 : 1;
        }

    if(current_fmt == 2){

        smatch search;

        if(!regex_search(line,  search, regex("^[\\*\\+]+"))){
            if(regex_search(line, vul_id_regex)){
                return IdentifyFormat(0, line);
                }
            return -2;
            }

        line = line.substr(search.str().size());
        }

    return current_fmt;
    }

//-----------------------------------------------------------------------------
// Get full package version (version and release) from a given header of
// a changelog entry

string GetPackageVerRel(const string &line){

    // The pos var is needed to skip spaces at the right (after the version and
    // release part)
    size_t pos = line.size();

    while(pos > 0 && line[pos] == ' '){
        pos--;
        }

    if(pos == 0){
        return "";
        }

    return line.substr(line.find_last_of(' ', pos) + 1);
    }

//-----------------------------------------------------------------------------
// Get a set of vulnerability IDs from a given line and cut a processed part
// of that line

set<string> GetIDsAndCutProcessedPart(string &line){

    #ifdef LCVEMANAGER_FULL_DEBUGGING
    cout << "\t[GetIDsAndCutProcessedPart is called]" << endl
        << "\tLine to process is \"" << line << "\"" << endl;
    #endif

    set<string> vul_ids;

    while(true){

        if(line.empty()){
            break;
            }

        smatch match;

        if(!regex_search(line, match, vul_id_regex)){
            break;
            }

        #ifdef LCVEMANAGER_FULL_DEBUGGING
        cout << "\t\tGot match: {"
            << "\"" << match[0].str() << "\"; "
            << "\"" << match[1].str() << "\"}" << endl;
        #endif

        vul_ids.insert(match[1].str());
        line = line.substr(match.position(0) + match[0].str().size());

        #ifdef LCVEMANAGER_FULL_DEBUGGING
        cout << "\t\tNot processed part of the line is \"" << line << "\"" << endl;
        #endif
        }

    #ifdef LCVEMANAGER_FULL_DEBUGGING
    cout << "\t[GetIDsAndCutProcessedPart is done]" << endl;
    #endif

    return vul_ids;
    }

//-----------------------------------------------------------------------------
// Get IDs of vulnerabilities from a given multiline text

pair<map<string, set<string>>, string> GetFixedVerAndVulIDs(const string &text){

    #ifdef LCVEMANAGER_FULL_DEBUGGING
    cout << "[GetFixedVerAndVulIDs is called]" << endl
        << "Text to process is \"" << text << "\"" << endl;
    #endif

    string package_ver;
    map<string, set<string>> fixed_versions_and_vul_ids;
    string warn;

    int fmt = 0;
    int count = 0;

    string line;
    stringstream ss(text, ios_base::in);

    // Iterating through all lines of a given string
    while(getline(ss, line, '\n')){

        #ifdef LCVEMANAGER_FULL_DEBUGGING
        cout << "Current line is \"" << line << "\"" << endl;
        cout << "\tCurrent fmt is " << fmt << endl;
        #endif

        count++;

        if(line.empty() || line.front() == '*'){
            if(line.front() == '*'){
                package_ver = GetPackageVerRel(line);
                #ifdef LCVEMANAGER_FULL_DEBUGGING
                cout << "\tThis is a header of a changelog entry" << endl;
                     << "\t\tPackage version is " << package_ver << endl;
                #endif
                }
            fmt = 0;
            continue;
            }

        Trim(line);
        line = ToLower(line);

        if(line.empty() || line.front() == '-'){
            fmt = 0;
            if(line.empty()){
                continue;
                }
            }

        #ifdef LCVEMANAGER_FULL_DEBUGGING
        cout << "\tPreprocessed line is \"" << line << "\"" << endl;
        #endif

        while(true){

            fmt = IdentifyFormat(fmt, line);

            #ifdef LCVEMANAGER_FULL_DEBUGGING
            cout << "\tIdentified fmt is " << fmt << endl;
            #endif

            if(fmt <= 0){
                fmt = abs(fmt);
                break;
                }

            // Adding IDs found in a current line
            for(const auto &vul_id : GetIDsAndCutProcessedPart(line)){
                #ifdef LCVEMANAGER_FULL_DEBUGGING
                cout << "\t\tGot ID: " << vul_id << endl;
                #endif
                fixed_versions_and_vul_ids[package_ver].insert(vul_id);
                }

            #ifdef LCVEMANAGER_FULL_DEBUGGING
            cout << "\tNot processed part of a preprocessed line is "
                << "\"" << line << "\"" << endl;
            #endif

            if(fmt == 1 && !line.empty()){
                fmt = 0;
                continue;
                }

            break;
            }
        }

    #ifdef LCVEMANAGER_FULL_DEBUGGING
    cout << "[GetFixedVerAndVulIDs call is done]" << endl;
    #endif

    return {fixed_versions_and_vul_ids, warn};
    }

//-----------------------------------------------------------------------------
